#ifndef FACETTES_SYSTEME_H
#define FACETTES_SYSTEME_H

#include "Facette.h"
#include <map>
#include <algorithm>
#include <memory>

class FacettesSysteme {
   std::map<Facette::id, std::unique_ptr<Facette>> facettes;
   FacettesSysteme();
public:
   FacettesSysteme(const FacettesSysteme&) = delete;
   FacettesSysteme& operator=(const FacettesSysteme&) = delete;
   //
   // Installer une facette signifie l'integrer au serveur de facettes.
   // l'operation est O(log n) avec un certain effet de bord. On triche
   // ici en modifiant it->second parce que it->first ne change pas
   // (on n'endommage pas l'arbre!)
   //
   template <class F>
      void installer(std::unique_ptr<Facette> p) {
         using namespace std;
         if (auto it = facettes.find(F::id); it != end(facettes))
            swap(p, it->second);
         else
            facettes[F::id] = std::move(p);
      }
   //
   // Obtenir une facette par son type... O(log n)
   //
   template <class F>
      const F& getFacette() const {
         return *static_cast<const F*>(const_cast<FacettesSysteme&>(*this).facettes[F::id].get());
      }
   static auto &get() {
      static FacettesSysteme singleton;
      return singleton;
   }
};


template <class F, class L>
   const F& utiliser_facette(const L &conteneur) {
      return conteneur.getFacette<F>();
   }

#endif
