#include "SequenceTaches.h"
//
// Coteux
//
#include "Etat.h"
#include <chrono>
#include <iostream>
#include <string>
#include <string_view>
#include <memory>
using namespace std;
using namespace std::chrono;

int SequenceTaches::moins_charge(int profondeur) {
   auto minpos = NDX_ENFANT_MIN;
   if (profondeur == 0) {
      for (int i = NDX_ENFANT_MIN; i <= NDX_ENFANT_MAX; ++i)
         if (enfants[minpos]->size() > enfants[i]->size())
            minpos = i;
   } else {
      auto minval = enfants[minpos]->moins_charge(profondeur - 1);
      for (int i = NDX_ENFANT_MIN; i <= NDX_ENFANT_MAX; ++i) {
         auto curval = enfants[i]->moins_charge(profondeur - 1);
         if (minval > curval) {
            minpos = i;
            minval = curval;
         }
      }
   }
   return static_cast<int>(minpos);
}

bool SequenceTaches::executer(delai_t alloue) {
   auto &exec = EtatGlobal.etats_executif;
   bool fini = false;
   const auto avant = high_resolution_clock::now();
   auto debut = taches.begin_from(depart),
        fin = taches.end_from(depart);
   auto ecoule = high_resolution_clock::now() - avant;
   for (; debut != fin && ecoule < alloue; ++debut) {
      auto && f = *debut;
      f();
      ecoule = high_resolution_clock::now() - avant;
   }
   ++exec.n_it_sequence[no_sequence];
   exec.charge_totale_sequence[no_sequence] += taches.size ();
   exec.charge_moyenne_sequence[no_sequence] =
      static_cast<double>(exec.charge_totale_sequence[no_sequence]) /
      static_cast<double>(exec.n_it_sequence[no_sequence]);
   if (ecoule < alloue) {
      auto restant = duration_cast<milliseconds>(alloue - ecoule);
      exec.temps_restant_total_sequence[no_sequence] += restant.count();
      exec.temps_restant_moyen_sequence[no_sequence] =
         exec.temps_restant_total_sequence[no_sequence] / exec.n_it_sequence[no_sequence];
      fini = true;
      depart = begin(taches);
      if (enfants[enfant_cur]) {
         fini = enfants[enfant_cur]->executer(restant);
         prochain_enfant();
      }
   } else {
      ++exec.n_debordements_sequence[no_sequence];
   }
   return fini;
}

//
// Retourne le nombre de sequences creees
//
int SequenceTaches::creer_branches(int niveaux) {
   int nbranches = 0;
   --niveaux;
   for (int i = NDX_ENFANT_MIN; i <= NDX_ENFANT_MAX; ++i) {
      ++nbranches;
      enfants[i] = make_unique<SequenceTaches>(no_sequence + nbranches);
      if (niveaux)
         nbranches += enfants[i]->creer_branches(niveaux);
   }
   depart = begin(taches);
   return nbranches;
}

void SequenceTaches::afficher_stats(ostream &os, int niveau) {
   auto &exec = EtatGlobal.etats_executif;
   static constexpr auto PREFIXE_INDENTATION = "   "sv;
   n_times(niveau, [&os] { os << PREFIXE_INDENTATION; });
   os << "Sequence " << no_sequence
      << ", executee " << exec.n_it_sequence[no_sequence] << " fois"
      << ", debordee " << exec.n_debordements_sequence[no_sequence] << " fois"
      << endl;
   n_times(niveau, [&os] { os << PREFIXE_INDENTATION; });
   os << "-> Unites de temps libres (moyenne): "
      << exec.temps_restant_moyen_sequence[no_sequence]
      << endl;
   n_times(niveau, [&os] { os << PREFIXE_INDENTATION; });
   os << "-> Charge moyenne de " << exec.charge_moyenne_sequence[no_sequence] << " taches"
      << endl;
   for (int i = NDX_ENFANT_MIN; i <= NDX_ENFANT_MAX; ++i)
      if (enfants[i])
         enfants[i]->afficher_stats(os, niveau + 1);
}

void SequenceTaches::detruire_branches() {
   for (int i = NDX_ENFANT_MIN; i <= NDX_ENFANT_MAX; ++i)
      if (enfants[i])
         enfants[i]->detruire_branches();
}
