#ifndef SEQUENCE_TACHES_H
#define SEQUENCE_TACHES_H

#include "file_circulaire.h"
#include "Tache.h"
#include "Utilitaires.h"
#include "observer_ptr.h"
#include <iosfwd>
#include <memory>
#include <chrono>
#include <functional>

//
// Une sequence de taches regroupe des taches qui seront
// executees cycliquement et en sequence et qui ont la
// particularite d'tre plus importantes que leurs enfants.
//
// Une sequence de taches peut s'interrompre. Dans ce cas,
// elle reprendra o elle etait rendue lors de la prochaine
// invocation de ses services
//
// On peut bouger une tache vers le haut (donc augmenter sa
// priorite) ou vers le bas (donc reduire sa priorite), donc
// les taches seront enregistrees ailleurs
//
class SequenceTaches {
public:
   using delai_t = std::chrono::high_resolution_clock::duration;
private:
   static constexpr int
      NDX_ENFANT_MIN = 0,
      NDX_ENFANT_MAX = 1,
      NB_ENFANTS = NDX_ENFANT_MAX - NDX_ENFANT_MIN + 1;
   int enfant_cur{ NDX_ENFANT_MIN };
   observer_ptr<SequenceTaches> parent;
   std::unique_ptr<SequenceTaches> enfants[NB_ENFANTS];
   file_circulaire<std::function<void()>> taches;
   file_circulaire<std::function<void()>>::iterator depart;
   int no_sequence;
   void prochain_enfant() {
      enfant_cur = prochain_circulaire(enfant_cur, NDX_ENFANT_MIN, NDX_ENFANT_MAX);
   }
   int moins_charge(int profondeur);
   void detruire_branches();
public:
   auto size() const noexcept {
      return taches.size();
   }
   class IndiceIllegal {};
   SequenceTaches(int no_sequence, observer_ptr<SequenceTaches> parent = nullptr)
      : no_sequence{ no_sequence }, parent{ parent } {
      using namespace std;
      depart = end(taches);
   }
   //
   // Retourne le nombre de sequences creees
   //
   int creer_branches(int nb_niveaux);
   template <class F>
      void ajouter(F f, int profondeur) {
         using namespace std;
         if(profondeur == 0) {
            taches.push_back(f);
            depart = begin(taches);
         } else {
            const int profondeur_candidate = profondeur - 1;
            auto ndx = moins_charge(profondeur_candidate);
            enfants[ndx]->ajouter(f, profondeur_candidate);
         }
      }
   bool executer(delai_t alloue);
   ~SequenceTaches() {
      detruire_branches();
   }
   void afficher_stats(std::ostream&, int niveau = 0);
};

#endif
