#ifndef OBSERVER_PTR_H
#define OBSERVER_PTR_H

template <class T>
   class observer_ptr {
      T *p;
   public:
      constexpr observer_ptr(T *p) noexcept : p{ p } {
      }
      constexpr T& operator*() noexcept {
         return *p;
      }
      constexpr const T& operator*() const noexcept {
         return *p;
      }
      constexpr T* operator->() noexcept {
         return p;
      }
      constexpr const T* operator->() const noexcept {
         return p;
      }
      constexpr bool operator==(const observer_ptr &autre) const {
         return p == autre.p;
      }
      constexpr bool operator!=(const observer_ptr &autre) const {
         return !(*this == autre);
      }
      template <class U>
         constexpr bool operator==(const observer_ptr<U> &autre) const {
            return p == &*autre;
         }
      template <class U>
         constexpr bool operator!=(const observer_ptr<U> &autre) const {
            return !(*this == autre);
         }
      template <class U>
         constexpr bool operator==(const U *q) const {
            return p == q;
         }
      template <class U>
         constexpr bool operator!=(const U *q) const {
            return !(*this == q);
         }
      constexpr T *get() const noexcept {
         return p;
      }
      constexpr bool empty() const noexcept {
         return !p;
      }
      constexpr explicit operator bool () const noexcept {
         return !empty();
      }
   };

template <class T>
      constexpr observer_ptr<T> observe(T && p) {
         static_assert(std::is_pointer_v<T>);
         return { p };
      }

#endif
